/*
 * Implement the Usenet control messages, as per RFC 1036 (nee 850).
 * These are fairly infrequent and can afford to be done by
 * separate programs.  They are:
 *
 * cancel message-ID	restricted to Sender: else From: or root, in theory
 * ihave message-ID-list remotesys	generate a sendme from message-ID-list
 * sendme message-ID-list remotesys	send articles named to remotesys
 * (ihave/sendme is semi-documented in the RFCs, kludgey and broken
 * in B2.10.)
 *
 * newgroup groupname	must be Approved:
 * rmgroup groupname	must be Approved:; allow some local control over this
 * sendsys				mail to Reply-To: else From:
 * senduuname			        ditto
 * version				ditto
 */

#include <stdio.h>
#include <ctype.h>
#ifndef AMIGA
#  include <sys/types.h>
#endif /* AMIGA */

#include "libc.h"
#include "news.h"
#include "config.h"
#include "headers.h"
#include "article.h"
#include "caches.h"
#include "history.h"

#define NO_FILES ""
#define SUBDIR binfile("ctl")		/* holds shell scripts */


/* imports from news */
extern statust snufffiles(); 
extern void ihave(), sendme();

/* forwards */
FORWARD statust cancelart();
FORWARD void runctlmsg();

/*
 * Implement control message specified in "art".
 * Because newgroup and rmgroup may modify the active file, for example,
 * we must flush in-core caches to disk first and reload them afterward.
 * We handle cancels in this process for speed and dbm access.
 * We handle ihave & sendme in this process for dbm access and
 * to work around syntax restrictions (<>).
 *
 * In future, one could pass header values to scripts as arguments or
 * in environment, as NEWS* variables, to save time in the scripts.
 */

void ctlmsg(art)
struct article *art;
{
	int pid, deadpid;
	int wstatus;
	char *inname = art->a_tmpf, *ctlcmd = art->h.h_ctlcmd;
	static char nmcancel[] = "cancel ";
	static char nmihave[] = "ihave ";
	static char nmsendme[] = "sendme ";

	if (STREQN(ctlcmd, nmcancel, STRLEN(nmcancel))) {
		art->a_status |= cancelart(ctlcmd + STRLEN(nmcancel));
		return;
	}
	if (STREQN(ctlcmd, nmihave, STRLEN(nmihave))) {
		ihave(ctlcmd + STRLEN(nmihave), art);
		return;
	}
	if (STREQN(ctlcmd, nmsendme, STRLEN(nmsendme))) {
		sendme(ctlcmd + STRLEN(nmsendme), art);
		return;
	}

	/*
	 *	These "other" ctlmsgs are just plain not handled on an Amiga.
	 *	I did this primarily so that I wouldn't have to write the
	 *	fork() and system() code that Manx left out of the library.
	 *	What a bunch of dildoes!!
	 *
	 *	I took this #ifdef out since I'm going to write a stub for
	 *	the fork(), system(), popen(), and pclose() routines so that
	 *	I can determine how these routines need to be implemented (at
	 *	least, implemented here).
	 */

	art->a_status |= synccaches();	/* 'active' file must be closed! -IF- */
	(void) fflush(stdout);		/* cautious */
	(void) fflush(stderr);		/* cautious */

	runctlmsg(ctlcmd, inname);	/* Now lets execute the controlcmd */
	art->a_status |= loadcaches();	/* And reopened! -IF- */
}

/*
 * In theory (RFC 1036 nee 850), we should verify that the user issuing
 * the cancel (the Sender: of this article or From: if no Sender) is the
 * Sender: or From: of the original article or the local super-user.
 *
 * In practice, this is a lot of work and since anyone can forge news
 * (and thus cancel anything), not worth the effort.
 *
 * Ignore ST_ACCESS while cancelling an already-seen article since the
 * article may have been cancelled before or may have a fake history entry
 * because the cancel arrived before the article.
 *
 * If the article being cancelled has not been seen yet, generate a history
 * file entry for the cancelled article in case it arrives after the cancel
 * control.  The history file entry will cause the cancelled article to be
 * rejected as a duplicate.
 */

STATIC statust cancelart(msgidstr)
char *msgidstr;
{
	register char *wsp;
	register char *msgid = strsave(msgidstr);
	register int idbytes;
	register char *wholemsgid = msgid;
	register statust status = ST_OKAY;

	/* skip leading whitespace in msgid */
	while (*msgid != '\0' && isascii(*msgid) && isspace(*msgid))
		++msgid;
	/* replace trailing whitespace with NULs; `wsp >= msgid' is not safe */
	idbytes = strlen(msgid);
	for (wsp = msgid + idbytes - 1; idbytes-- > 0 &&
					    isascii(*wsp) && isspace(*wsp); --wsp)
		*wsp = '\0';

	if (alreadyseen(msgid)) {
		register char *histent, *filelist;

		histent = gethistory(msgid);
		if (histent != NULL && (filelist = findfiles(histent)) != NULL)
			status |= snufffiles(filelist) & ~ST_ACCESS;
	} else {
		status |= fakehist(msgid, DEFEXP, NO_FILES);	/* start log */
		(void) putchar('\n');		/* end log line */
	}
	free(wholemsgid);
	return status;
}

/*
 * Execute a non-builtin control message by searching $NEWSCTL/bin and
 * $NEWSBIN/ctl for the command named by the control message.
 * runctlmsg is called from a child of relaynews, so it must always
 * call _exit() rather than exit() to avoid flushing stdio buffers.
 *
 * Enforce at least minimal security: the environment was standardised at
 * startup, including PATH and IFS; close non-standard file descriptors;
 * reject shell metacharacters in ctlcmd.
 */

STATIC void runctlmsg(ctlcmd, inname)			/* child process */
register char *ctlcmd, *inname;
{
	char cmd[256];
        char *tmp = strsave(ctlcmd);
        char *ptr;

        if((ptr = strchr(tmp, ' ')) != NULL)
          *ptr = '\0'; /* Only command, not parameter */

	(void) strcpy(cmd, ctlfile("bin"));
	(void) strcat(cmd,"/newscontrol ");
	(void) strcat(cmd, tmp);
	(void) strcat(cmd, " ");
	(void) strcat(cmd, inname);

	(void)Execute(cmd, 0, 0); /* Execute() returns not the right value */
}
